/* ============================================================================
 *                         t i m e r _ i n t e r f a c e . c
 * ============================================================================
 * Cytonome CAB 3.0 Cypress CY8C5868AXI-LP035 Firmware
 * Copyright Cytonome ST, LLC 2022
 * All Rights Reserved
 *
 * UNPUBLISHED, LICENSED SOFTWARE.
 *
 * CONFIDENTIAL AND PROPRIETARY INFORMATION
 * WHICH IS THE PROPERTY OF Cytonome ST, LLC.
 *
 * This firmware started out as 'NextGenFirmware' 'CAB_REV_A2' from 2021 or earlier, which was intended for CAB 2.x, probably the later version
 * It has been **extensively** modified for CAB 3.0 Rev B17
 *
 * Revision History:
 *  001 8 Jun 22 <sek> updated notes
 *  002 14 Jul 22 <sek> GetmSCount()
 *  003 21 Jul 22 <sek> SysTimers code
 *  004 27 Jul 22 <sek> Duration is uint
 *  005 29 Jul 22 <sek> systimer_pause(), systimer_resume(), added active to systimer structure
 *  006 02 Aug 22 <sek> additional tests for systimer_pause(), systimer_resume()
 *  007 24 Jan 23 <sek> ARRAY markers
 *  008 25 Jan 23 <sek> CPU Usage markers, changed CySysTickInit() to CySysTickStart()
 *
* ============================================================================
*/

/*
* Timer interface Hardware
    Fan PWM, externally buffered, drives all 6 fans
    Fan speed measurement

Timer Interface Firmware
    SysTick
        SysTickTimersInitialize()
        SysTickSecondsCounter()
        GetSysTickSeconds()
        Are we doing anything that requires resolution better than 1S?
    Hardware timers require control code:
        Fan PWM
            limit range of settings
        Fan Tach measurement
            Is this a noisy signal?
            average multiple cycles to reduce noise?
        Arc Detection
            Arcs per second reset
            Arc detected reset
*/

#include "gsfstd.h"

#include "main.h"
#include <assert.h>
#include "debug.h"
#include "timer_interface.h"

#define CPU_USAGE   1          // set to 0 through 10 to measure execution time of various blocks of code
#define START_CPU_USAGE(x)      if (x == CPU_USAGE) SPARE_2_Write(1);
#define END_CPU_USAGE(x)        if (x == CPU_USAGE) SPARE_2_Write(0);
#define KEEPALIVE_AVAILABLE


enum tagThermalControlErrors
{
	TIMER_ERROR_NONE = TIMER_ERROR_BASE,
	TIMER_ERROR_UNEXPECTED_TICK,		// 1 unexpected timer tick event
	TIMER_ERROR_UNEXPECTED_EVENT,		// 2 unexpected event
	TIMER_ERROR_INVALID_STATE,		    // 3 not a valid state
	TIMER_ERROR_INVALID_SUBSTATE,		// 4 not a valid state
	TIMER_ERROR_INVALID_TIMER,	        // 5 invalid duty cycle
    TIMER_ERROR_TIMER_NOT_PAUSED,       // 6 check for systimer_resume()
    TIMER_ERROR_TIMER_NOT_RUNNING,      // 7 check for systimer_pause()

	TIMER_ERROR_UNPROCESSED_EVENT = TIMER_ERROR_BASE + 0x0F
};


// forward references
//  initialize the timer subsystem
void systimer_init(void);
//  service the timers.
//  This function should be called by a periodic 1msec interrupt.
void systimer_isr(void);

//
// Timer functions.  
// In this implementation the 1mS SysTick counter is used
// to generate a seconds counter.
//
// NOTE: Starting the saved seconds count at 1 so that it returns something
// other than zero if it's used at system-startup before 1 second has expired.
// 
volatile static uint32 s_SysTickInSeconds = 1;

PRIVATE_INIT int16 mSCount = 0;

//============================================================================
// definition of a timer object
typedef struct 
{
    uint msec;       // counter to track current timer value
    uint duration;   // duration of timer to expiration
    BOOL active;     // flag to mark active or inactive; required for pause and resume
} systimer_t;

// Array containing all the timers in the system.
ARRAY systimer_t systemTimers[SYSTIMER_MAX];

//============================================================================
// Initialize the array of systemTimers
//
void systimer_init(void)
{
    // clear all timers
    for (int i = 0; i < SYSTIMER_MAX; i++)
    {
        systemTimers[i].msec = 0;
        systemTimers[i].duration = 0;
        systemTimers[i].active = FALSE;
        
    }
}


//============================================================================
// systimer 1mS interrupt code.
// bumps timers, and checks for expired
//
void systimer_isr(void)
{

    START_CPU_USAGE(1);
    
    // check all timers; only increment timers that have not expired
    for (int i = 0; i < SYSTIMER_MAX; i++)
    {
        if (systemTimers[i].msec < systemTimers[i].duration)
        {
            systemTimers[i].msec++;
        }
        else if (systemTimers[i].duration > 0)
        {
            // a timer has expired, so we must have something to do!
           ;
        }
    }
    
    END_CPU_USAGE(1);
    
}

//============================================================================
// test for a specific systimer timed out
//
BOOL systimer_expired(uint8 timerIndex)
{
    if (timerIndex >= SYSTIMER_MAX)
    {
        RuntimeError(TIMER_ERROR_INVALID_TIMER);
        return FALSE;
    }

    BOOL bRetVal = FALSE;
    
    // if msec counter exceeds the timer duration, the timer has expired
    bRetVal = systemTimers[timerIndex].msec >= systemTimers[timerIndex].duration;
    
    if (bRetVal IS_TRUE)
    {
        systemTimers[timerIndex].active = FALSE;
    }
    
    return bRetVal;

}


//============================================================================
// start a systimer
//
void systimer_start(uint8 timerIndex, uint16 duration)
{
    if (timerIndex >= SYSTIMER_MAX)
    {
        RuntimeError(TIMER_ERROR_INVALID_TIMER);
        return;
    }

    systemTimers[timerIndex].msec = 0;

    // add one to the desired expiration to ensure the routine waits for at least
    // the time desired. A timer set for a duration of X will expire somewhere between 
    // X and X+1 milliseconds 
    systemTimers[timerIndex].duration = duration + 1;
    
    systemTimers[timerIndex].active = TRUE;
}

//============================================================================
// pause a systimer
//
void systimer_pause(uint8 timerIndex)
{
    if (timerIndex >= SYSTIMER_MAX)
    {
        RuntimeError(TIMER_ERROR_INVALID_TIMER);
        return;
    }

    if (systemTimers[timerIndex].msec IS 0)
    {
        RuntimeError(TIMER_ERROR_TIMER_NOT_RUNNING);
        return;
    }
    
    systemTimers[timerIndex].active = FALSE;
  
}

//============================================================================
// resume a paused systimer
//
void systimer_resume(uint8 timerIndex)
{
    if (timerIndex >= SYSTIMER_MAX)
    {
        RuntimeError(TIMER_ERROR_INVALID_TIMER);
        return;
    }
    
    if ((systemTimers[timerIndex].active IS FALSE)
        AND (systemTimers[timerIndex].msec IS_NOT 0))
    {
        systemTimers[timerIndex].active = TRUE;
    }
    else
    {
        RuntimeError(TIMER_ERROR_TIMER_NOT_PAUSED);
        return;
    }
  
}

//============================================================================
// stop a systimer when it is no longer needed
//
void systimer_stop(uint8 timerIndex)
{
    if (timerIndex >= SYSTIMER_MAX)
    {
        RuntimeError(TIMER_ERROR_INVALID_TIMER);
        return;
    }

    systemTimers[timerIndex].msec = 0;
    systemTimers[timerIndex].duration = 0;
    systemTimers[timerIndex].active = FALSE;
}

//============================================================================
//
// this is the system level mS ISR
//
static void SysTickmilliSecond_ISR(void)
{
    START_CPU_USAGE(2);
    //  Executes from ISR context.  
    //  Attached to the 1mS SysTickTimer. 
    //  Increments a count in seconds.
    //

    if ( (++mSCount % 1000) == 0 )
    {
        s_SysTickInSeconds++;
        mSCount = 0;
    }
    
    // bump all active systimers
    systimer_isr();
    
    END_CPU_USAGE(2);
}

// return the current mS counter value
int16 GetmSCount(void)
{
    return (mSCount);
}

//============================================================================
//
volatile uint32 GetSysTicksInSeconds(void)
{
    // Public function, 
    // returns system tick count in seconds.
    // (Used when timeouts are defined in seconds).
    //
    return s_SysTickInSeconds;
}

//============================================================================
//
void SysTickTimersInitialize(void)
{
    //  Initialize the 1mS systick counter
    // 
    CySysTickStart();

    // initialize all of the systimers
    systimer_init();
    
    //  Attach the callback that counts seconds.
    // 
    CySysTickSetCallback(ISR_PRIORITY_0, SysTickmilliSecond_ISR);

    //  Enable the systick counter.
    //
    CySysTickEnable();
}



